from re import match
from copy import copy
from warnings import warn
from getpass import getpass
from fqdn import FQDN
from docksible.constants import *

INSECURE_CLI_PASS_WARNING = 'It\'s insecure to pass passwords via CLI args! If you are sure that you want to do this, rerun this command with the --insecure-cli-password flag.'

class ArgValidator():

    def __init__(self, args):
        self.raw_args       = args
        self.validated_args = args


    def get_raw_args(self):
        return self.raw_args


    def get_validated_args(self):
        return self.validated_args


    # TODO?
    def handle_defaults(
        self,
        default_args,
        ask_user=False,
        verbose=False
    ):
        """Handles defaults in various cases, optionally setting values with
        application wide defaults or overriding values, and optionally
        printing warnings.

        Positional arguments

        default_args -- A list of dictionaries, which defines the arguments,
                        and how to treat them. The following dictionary serves
                        as an example:
            {
                # required
                'arg_name': 'some_cli_arg', # The user passed this in as
                                            # --some-cli-arg - We also have a
                                            # helper function to get the arg
                                            # name in that format.
                # required
                'cli_default_value': DEFAULT_ARG_VALUE,
                # optional
                'override_default_value': 'some use case specific default',
            }

        ask_user -- Optional. If True, then if we got default values,
                    prompt the user to input their own value,
                    and if they leave it blank,
                    fall back to default values. Defaults to False.

        verbose --  Optional. If True, then if we are using some
                    default value, warn the user about this. This is useful
                    for credentials, in case we're falling back to some
                    insecure value.
        """
        for arg_dict in default_args:
            try:
                default_value = arg_dict['override_default_value']
            except KeyError:
                default_value = arg_dict['cli_default_value']
            user_value = getattr(self.raw_args, arg_dict['arg_name'])

            if user_value == arg_dict['cli_default_value']:
                if ask_user:
                    tmp_val = input(
                        'Got no {}. Please enter a value now, or leave blank to default to \'{}\': '.format(

                            self.var_name_to_cli_arg(arg_dict['arg_name']),
                            default_value
                        )
                    )
                    if tmp_val == '':
                        tmp_val = default_value
                    default_value = tmp_val

                setattr(
                    self.validated_args,
                    arg_dict['arg_name'],
                    default_value
                )

                if verbose:
                    print('Using {} value \'{}\'.'.format(
                        self.var_name_to_cli_arg(arg_dict['arg_name']),
                        default_value
                    ))


    def var_name_to_cli_arg(self, var_name):
        return '--{}'.format(var_name.replace('_', '-'))


    def get_pass_and_check(self, prompt, min_length=0, confirm=False):
        password = getpass(prompt)
        while min_length > 0 and len(password) < min_length:
            password = getpass('That password is too short. Please enter another password: ')
        if confirm:
            double_check = getpass('Please retype password: ')
            if password == double_check:
                return password
            else:
                print('Passwords don\'t match. Please try again.')
                return self.get_pass_and_check(prompt, min_length, True)
        else:
            return password


    def validate_ansible_runner_args(self):
        try:
            user_at_host_split = self.raw_args.user_at_host.split('@')
            assert len(user_at_host_split) == 2
            self.validated_args.user = user_at_host_split[0]
            self.validated_args.host = user_at_host_split[1]
        except AssertionError:
            print("FATAL! First positional argument must be in the format 'user@host'.")
            return 1
        if self.raw_args.remote_sudo_password \
            and not self.raw_args.insecure_cli_password:
            print(INSECURE_CLI_PASS_WARNING)
            return 1
        if self.raw_args.ask_remote_sudo:
            self.validated_args.remote_sudo_password = self.get_pass_and_check(
                'Please enter sudo password for remote host: ')
        return 0


    def validate_database_args(self):

        if not self.raw_args.insecure_cli_password \
            and (self.raw_args.database_password \
                or self.raw_args.database_root_password
            ):

            print(INSECURE_CLI_PASS_WARNING)
            return 1

        # TODO: Add some option like --wordpress-defaults, to improve user
        # experience. Otherwise, the user would always be asked about defaulting
        # to 'wordpress' and 'wp_' for database name and table prefix, which
        # might be a little annoying.
        if self.raw_args.action == 'wordpress':
            self.handle_defaults([
                {
                    'arg_name': 'database_name',
                    'cli_default_value': None,
                    'override_default_value': 'wordpress',
                },
                {
                    'arg_name': 'database_username',
                    'cli_default_value': None,
                    'override_default_value': DEFAULT_DATABASE_USERNAME,
                },
                {
                    'arg_name': 'database_table_prefix',
                    'cli_default_value': DEFAULT_DATABASE_TABLE_PREFIX,
                    'override_default_value': 'wp_',
                },
            ], True, True)

        if not self.raw_args.database_root_password:

            self.validated_args.database_root_password = self.get_pass_and_check(
                'Please enter a database root password: ',
                0,
                True
            )

        if self.raw_args.action == 'backup':
            self.handle_defaults([
                {
                    'arg_name': 'database_name',
                    'cli_default_value': None,
                    'override_default_value': DEFAULT_DATABASE_NAME,
                },
                {
                    'arg_name': 'database_username',
                    'cli_default_value': None,
                    'override_default_value': DEFAULT_DATABASE_USERNAME,
                },
            ], True, True)

        if self.raw_args.database_username and not self.raw_args.database_password:
            self.validated_args.database_password = self.get_pass_and_check(
                'Please enter a database password: ',
                0,
                True
            )

        return 0


    def validate_ssl_args(self):

        if not self.raw_args.letsencrypt:
            return 0

        if self.raw_args.domain is None:
            self.validated_args.domain = self.validated_args.host

        while self.validated_args.email is None \
                or not match(r"[^@]+@[^@]+\.[^@]+", self.validated_args.email
                ):
            self.validated_args.email = input('Please enter a valid email address: ')

        return 0


    def validate_args(self):
        validate_methods = [
            'validate_ansible_runner_args',
            'validate_database_args',
            'validate_ssl_args',
        ]
        for method_name in validate_methods:
            method = getattr(self, method_name)
            result = method()
            if result != 0:
                return result

        return 0



import unittest
from docksible.docksible import Docksible


#######################################################
# TODO: Unit tests would be really cool.
# But it's a little difficult, because we need to mock
# some kind of server against which Docksible can run.
#######################################################

class TestDocksible(unittest.TestCase):

    def setUp(self):
        self.docksible = Docksible(
            user='user',
            host='localhost',
            action='wordpress',
            private_data_dir=os.path.join(
                'test',
                'tmp-private-data',
            )
        )


    def tearDown(self):
        self.docksible.cleanup_private_data()



from ansible_runner import Runner, RunnerConfig
from ansible_directory_helper.private_data import PrivateData
from .constants import *


class Docksible:

    def __init__(self, user, host, action,
            private_data_dir=DEFAULT_PRIVATE_DATA_DIR,
            database_root_password=None, database_username=None,
            database_password=None, database_name=None,
            letsencrypt=False,
            wordpress_auth_vars=None,
            domain=None, email=None,
            test_cert=False
        ):
        self.user   = user
        self.host   = host
        self.action = action

        self.private_data_helper = PrivateData(private_data_dir)
        self._init_inventory()

        self.runner_config = RunnerConfig(
            private_data_dir=private_data_dir,
            project_dir=PROJECT_DIR
        )
        self.runner = Runner(self.runner_config)

        self.database_root_password = database_root_password
        self.database_username = database_username
        self.database_password = database_password
        self.database_name = database_name
        self.letsencrypt = letsencrypt
        self.wordpress_auth_vars = wordpress_auth_vars
        self.domain = domain
        self.email = email
        self.test_cert = test_cert


    def _init_inventory(self):
        self.private_data_helper.add_inventory_groups('all')
        self.private_data_helper.add_inventory_host(self.host, 'all')
        self.private_data_helper.set_inventory_ansible_user(self.host, self.user)
        self.private_data_helper.write_inventory()


    def _update_env(self):
        extravars = [
            'database_root_password',
            'database_username',
            'database_password',
            'database_name',
            'wordpress_auth_vars',
            'domain',
            'email',
            'service_to_encrypt',
            'test_cert',
        ]
        for varname in extravars:
            try:
                if varname == 'service_to_encrypt':
                    value = self.action
                    if value == 'redmine':
                        self.private_data_helper.set_extravar('port_to_encrypt', 3000)
                    else:
                        self.private_data_helper.set_extravar('port_to_encrypt', 80)
                elif varname == 'test_cert':
                    value = self.get_certbot_test_cert_string()
                elif varname == 'domain':
                    if not self.domain:
                        value = self.host
                    else:
                        value = self.domain
                else:
                    value = getattr(self, varname)

                self.private_data_helper.set_extravar(varname, value)
            except AttributeError:
                pass
        self.private_data_helper.write_env()


    def set_runner_config_var(self, var, val):
        setattr(self.runner_config, var, val)


    def set_playbook(self, playbook):
        self.set_runner_config_var('playbook', playbook)


    def _prepare_config(self):
        self.runner_config.prepare()


    def run(self):
        self._update_env()
        self._prepare_config()
        self.runner.run()

        if self.letsencrypt:
            self.set_playbook('letsencrypt.yml')
            self._update_env()
            self._prepare_config()
            self.runner.run()


    def cleanup_private_data(self):
        self.private_data_helper.cleanup_dir()


    # TODO: Looks like this isn't being used yet.
    def get_certbot_domains_string(self):
        try:
            return '-d {}'.format(' -d '.join(self.domains_for_ssl))
        except TypeError:
            return ''


    def get_certbot_test_cert_string(self):
        return '--test-cert' if self.test_cert else ''



import os
from .helpers import *

USER_HOME_DIR = os.path.expanduser('~')
PROJECT_DIR = find_package_project_dir()
DEFAULT_PRIVATE_DATA_DIR = os.path.join(USER_HOME_DIR, '.docksible')
DEFAULT_BACKUPS_DIR = os.path.join(USER_HOME_DIR, '.docksible-backups')
DEFAULT_DATABASE_USERNAME = 'db-username'
DEFAULT_DATABASE_NAME = 'db_name'
DEFAULT_DATABASE_TABLE_PREFIX = ''
